package com.webchat.sso.service;

import com.webchat.common.bean.APIResponseBean;
import com.webchat.common.bean.APIResponseBeanUtil;
import com.webchat.common.constants.WebConstant;
import com.webchat.common.enums.RedisKeyEnum;
import com.webchat.common.enums.RoleCodeEnum;
import com.webchat.common.enums.UserStatusEnum;
import com.webchat.common.exception.AuthException;
import com.webchat.common.exception.BusinessException;
import com.webchat.common.service.RedisService;
import com.webchat.common.util.IDGenerateUtil;
import com.webchat.common.util.MD5Utils;
import com.webchat.common.util.StringUtil;
import com.webchat.common.util.TransactionSyncManagerUtil;
import com.webchat.domain.vo.request.UserLoginInfoRequestVO;
import com.webchat.domain.vo.request.UserRegistryInfoRequestVO;
import com.webchat.domain.vo.response.UserBaseResponseVO;
import com.webchat.rmi.user.UserServiceClient;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import java.util.regex.Pattern;

@Service
public class SSOCoreService {

    @Autowired
    private UserServiceClient userServiceClient;

    @Autowired
    private RedisService redisService;

    /**
     * 登录授权码兑换次数限制
     */
    private static final int MAX_EXCHANGE_COUNT = 2;

    /**
     * 账号登录服务
     *
     * @param loginInfoRequest
     * @return
     */
    public String login(UserLoginInfoRequestVO loginInfoRequest) {
        // 1. 查询账号是否存在
        UserBaseResponseVO user = this.getUserBaseInfo(loginInfoRequest.getMobile());
        Assert.notNull(user, "账号不存在！");
        // 2. 校验账号密码
        String inputPassword = MD5Utils.md5(loginInfoRequest.getPassword().concat(WebConstant.MD5_SALT));
        if (!ObjectUtils.equals(inputPassword, user.getPassword())) {
            throw new AuthException("密码错误！");
        }
        // 3. 生成登录授权码
        return genLoginOauthCode(user.getUserId());
    }

    /**
     * 用户注册
     *
     * @param request
     * @return
     */
    // 这里其实不需要事务，因为想分享事务管理器的用法和场景，刻意添加了事务注解
    public boolean registry(UserRegistryInfoRequestVO request) {
        APIResponseBean<Boolean> responseBean = userServiceClient.registry(request);
        if (APIResponseBeanUtil.isOk(responseBean)) {
            return responseBean.getData();
        }
        return false;
    }

    /**
     * 判断用户是否存在
     * @param mobile
     * @return
     */
    private boolean hasUserInfo(String mobile) {
        UserBaseResponseVO userBaseResponseVO = getUserBaseInfo(mobile);
        return userBaseResponseVO != null;
    }

    private UserBaseResponseVO getUserBaseInfo(String mobile) {
        APIResponseBean<UserBaseResponseVO> responseMessage = userServiceClient.userBaseInfo(mobile);
        if (APIResponseBeanUtil.isOk(responseMessage)) {
            return responseMessage.getData();
        }
        throw new AuthException("用户不存在");
    }

    /**
     * 生成登录授权码
     *
     * @return
     */
    private String genLoginOauthCode(String userId) {
        // 生成登录授权码
        String oauthCode = IDGenerateUtil.uuid();
        String cacheKey = RedisKeyEnum.OAUTH_CODE_CACHE.getKey(oauthCode);
        redisService.set(cacheKey, userId, RedisKeyEnum.OAUTH_CODE_CACHE.getExpireTime());
        return oauthCode;
    }

    /**
     * 根据授权码获取用户ID
     *
     * @param oauthCode
     * @return
     */
    public String getUserIdByOauthCode(String oauthCode) {
        Long exchangeCount = this.getOauthCodeExchangeCountCache(oauthCode);
        if (exchangeCount >= MAX_EXCHANGE_COUNT) {
            throw new AuthException("授权码使用次数过多");
        }
        String cacheKey = RedisKeyEnum.OAUTH_CODE_CACHE.getKey(oauthCode);
        String userId = redisService.get(cacheKey);
        if (StringUtils.isNotBlank(userId)) {
            this.addExchangeCountCache(oauthCode);
            return userId;
        }
        throw new AuthException("授权码无效");
    }

    private void addExchangeCountCache(String oauthCode) {
        String cacheKey = RedisKeyEnum.OAUTH_CODE_GET_COUNT_CACHE.getKey(oauthCode);
        redisService.increx(cacheKey, RedisKeyEnum.OAUTH_CODE_GET_COUNT_CACHE.getExpireTime());
    }

    private Long getOauthCodeExchangeCountCache(String oauthCode) {
        String cacheKey = RedisKeyEnum.OAUTH_CODE_GET_COUNT_CACHE.getKey(oauthCode);
        String cache = redisService.get(cacheKey);
        return StringUtils.isBlank(cache) ? 0L : Long.valueOf(cache);
    }
}
